import os
import json
import yaml
import shutil
from copy import deepcopy
from datetime import datetime
from distutils import dir_util

from hpogrid.configuration import (ConfigurationBase, ModelConfiguration, 
                                   HPOConfiguration, GridConfiguration,
                                   SearchSpaceConfiguration)

from hpogrid.utils.helper import get_base_path, copytree

class ProjectConfiguration(ConfigurationBase):
    _CONFIG_TYPE_ = 'project'
    _CONFIG_DISPLAY_NAME_ = 'project'
    _DESCRIPTION_ = 'Manage a project for hyperparamter optimization'
    _LIST_COLUMNS_ = ['Project Title']
    _NAME_HELP_ = 'Name given to the project'
    _CONFIG_FORMAT_ = \
        {
            'scripts_path': {
                'abbr': 'p',
                'description': 'Path to the location of training scripts'+\
                               ' (or the directory containing the training scripts)',
                'required': True,
                'type': str
            },
            'model_config': {
                'abbr': 'm',
                'description': 'Name of the model configuration to use',
                'required': True,
                'type': str
            },
            'search_space': {
                'abbr': 's',
                'description': 'Name of the search space configuration to use',
                'required': True,
                'type':str
            },
            'hpo_config': {
                'abbr': 'o',
                'description': 'Name of the hpo configuration to use',
                'required': True,
                'type':str
            },
            'grid_config': {
                'abbr': 'g',
                'description': 'Name of the grid configuration to use',
                'required': True,
                'type':str
            }
        }
    _CONFIG_CLS_ = {
        'model_config': ModelConfiguration,
        'hpo_config': HPOConfiguration,
        'grid_config': GridConfiguration,
        'search_space': SearchSpaceConfiguration
    }
    
    def __init__(self):
        super().__init__()
        self._scripts_path = None
        
    @classmethod
    def get_config_dir(cls):
        base_path = get_base_path()
        return os.path.join(base_path, 'projects')
    
    
    @classmethod
    def get_project_dir(cls, project_name):
        project_dir = os.path.join(cls.get_config_dir(), project_name)
        return project_dir
    
    @classmethod
    def get_config_path(cls, project_name:str=None, extension:str='json'):
        """Returns the full path of a configuration file
        
        Args:
            project_name: str
                Name of the project
            extension: str, default='json'
                File extension for configuration file
        """
        
        project_dir = cls.get_project_dir(project_name)
        project_path = os.path.join(project_dir, 'config',
                                    'project_config.{}'.format(extension))
        return project_path
    
    @classmethod
    def get_scripts_dir(cls, project_name:str=None):
        """Returns the path to the training scripts of a project
        
        Args:
            project_name: str
                Name of the project
        """
        
        project_dir = cls.get_project_dir(project_name)
        scripts_path = os.path.join(project_dir, 'scripts')
        return scripts_path
    
    @classmethod
    def get_proj_config_dir(cls, project_name:str=None):
        """Returns the path to the configuration file of a project
        
        Args:
            project_name: str
                Name of the project
        """
        
        project_dir = cls.get_project_dir(project_name)
        project_config_dir = os.path.join(project_dir, 'config')
        return project_config_dir    
    
    def _validate_arguments(self, **args):
        args = super()._validate_arguments(**args)
        scripts_path = args.pop('scripts_path', None)
        if (scripts_path == ConfigurationBase._DEFAULT_OPT_ARG_):
            print('INFO: Path to training scripts is not specified. Skipping...')
            scripts_path = None
        elif (scripts_path is not None) and (not os.path.exists(scripts_path)):
            raise FileNotFoundError('Path to training scripts {} does not exist.'.format(scripts_path))
            
        self._scripts_path = scripts_path

        # check if input configuration files exist
        for config_type in self._CONFIG_CLS_:
            config_cls = self._CONFIG_CLS_[config_type]
            if (config_type in args) and (args[config_type] != ConfigurationBase._DEFAULT_OPT_ARG_):
                config_name = args[config_type]
                args[config_type] = config_cls.load(config_name)
            else:
                print('INFO: Path to {} configuration is not specified. '
                      'Skipping...'.format(config_cls._CONFIG_DISPLAY_NAME_))
        return args
    
    @classmethod
    def validate(cls, config):
        proj_config = deepcopy(config)
        for config_type in cls._CONFIG_CLS_:
            config_cls = cls._CONFIG_CLS_[config_type]
            proj_config[config_type] = config_cls.validate(config[config_type])
        return proj_config
    
    @classmethod
    def _validate(cls, config):
        print('INFO: Started validation process for project configurations.')
        result = cls.validate(config)
        print('Info: Validation complete.')
        return result
    
    def save(self, name=None, config=None, scripts_path=None, action='create'):        
        if (name is None) and (config is None):
            name = self._name
            config = self._config
        if name is None: 
            raise ValueError('configuration name undefined')
        config = {'project_name':name, **config}
        project_dir = self.get_project_dir(name)
        
        action_map = {
            'create': 'Creating',
            'recreate': 'Recreating',
            'update': 'Updating'
        }
        
        print('INFO: {} project "{}"'.format(action_map.get(action, action.title()), name))
        if (os.path.exists(project_dir)):
            if  action == 'create':
                print('ERROR: Project "{}" already exists. If you want to overwrite,'
                    ' use "recreate" or "update" action instead of "create".'.format(name))
                return None
            elif action == 'recreate':
                backup_dir = self.get_project_dir('backup')
                os.makedirs(backup_dir, exist_ok=True)
                timestamp = datetime.now().strftime("%Y%m%d%H%M%S")
                backup_proj_dir = os.path.join(backup_dir, '{}_{}'.format(name, timestamp))
                shutil.move(project_dir, backup_proj_dir)
                print('INFO: Recreating project. Original project moved to backup directory {}.'.format(
                    backup_proj_dir))
        
        # create directories if not already exist
        config_dir = self.get_proj_config_dir(name)
        scripts_dir = self.get_scripts_dir(name)
        if not os.path.exists(project_dir):
            os.makedirs(project_dir)
        if not os.path.exists(config_dir):
            os.makedirs(config_dir)
        if not os.path.exists(scripts_dir):
            os.makedirs(scripts_dir)
        
        if scripts_path is None:
            scripts_path = self._scripts_path
        
        # copy training scripts
        if (scripts_path is not None):
            if os.path.exists(scripts_path):
                # copy training scripts to the project directory
                print('INFO: Copying training scripts from {} to {}'.format(scripts_path, scripts_dir))
                # copy contents of directory to project/scrsipts/
                if os.path.isdir(scripts_path):
                    copytree(scripts_path, scripts_dir)
                else:
                    shutil.copy2(scripts_path, scripts_dir)
            else:
                raise FileNotFoundError('Path to training scripts {} does not exist.'.format(scripts_path))
        
        config_path = self.get_config_path(name)
        config_path_yaml = self.get_config_path(name, extension='yaml')
        if (os.path.exists(config_path)) and (action=='create'):
            print('ERROR: {} configuration "{}" already exists.'
                'If you want to overwrite, use "recreate" or "update" instead.'.format(
                self._CONFIG_DISPLAY_NAME_.title(), config_path))
        else:
            with open(config_path, 'w') as config_file:
                json.dump(config, config_file, indent=2)
            with open(config_path_yaml, 'w') as config_file_yaml:
                yaml.dump(config, config_file_yaml, default_flow_style=False, sort_keys=False)
            action_map = {'create': 'Created', 'recreate': 'Recreated', 'update': 'Updated'}
            print('INFO: {} {} configuration {}'.format(action_map[action], self._CONFIG_TYPE_, config_path))
            print('INFO: {} {} configuration {}'.format(action_map[action], self._CONFIG_TYPE_, config_path_yaml))
            self.show(name)


from hpogrid.configuration import ConfigurationBase
from hpogrid.components.defaults import (kSearchAlgorithms, kSchedulers, 
                                         kMetricMode, kDefaultSearchAlgorithm,
                                         kDefaultScheduler, kDefaultMetric,
                                         kDefaultMetricMode, kDefaultMaxConcurrent)

class HPOConfiguration(ConfigurationBase):
    _CONFIG_TYPE_ = 'hpo_config'
    _CONFIG_DISPLAY_NAME_ = 'HPO'
    _DESCRIPTION_ = 'Manage configuration for hyperparameter optimization'
    _LIST_COLUMNS_ = ['HPO Configuration']
    _CONFIG_FORMAT_ = \
        {
            'algorithm': {
                'abbr': 'a',
                'description': 'Algorithm for hyperparameter optimization',
                'required': False,
                'type': str,
                'choice': kSearchAlgorithms,
                'default': kDefaultSearchAlgorithm
            },
            'scheduler': {
                'abbr': 's',
                'description': 'Trial scheduling method for hyperparameter optimization',
                'required': False,
                'type': str,
                'choice': kSchedulers,
                'default': kDefaultScheduler
            },        
            'metric' : {
                'abbr': 'm',
                'description': 'Evaluation metric to be optimized',
                'required': False,
                'type': str,
                'default': kDefaultMetric
            },
            'extra_metrics': {
                'abbr': 'e',
                'description': 'Additional metrics to be saved during the training',
                'required': False,
                'type': (list, type(None)),
                'default': None
            },
            'mode' : {
                'abbr': 'o',
                'description': 'Mode of optimization (either "min" or "max")',
                'required': False,
                'type': str,
                'choice': kMetricMode,
                'default': kDefaultMetricMode
            },
            'resource': {
                'abbr': 'r',
                'description': 'Resource allocated to each trial',
                'required': False,
                'type': (dict, type(None)),
                'default': None
            },
            'num_trials': {
                'abbr': 'n',
                'description': 'Number of trials (search points)',
                'required': True,
                'type': int,
            },
            'log_dir': {
                'abbr': 'l',
                'description': 'Logging directory',
                'required': False,
                'type': (str, type(None)),
                'default': None
            },
            'verbose': {
                'abbr': 'v',
                'description': 'Verbosity level of Ray Tune',
                'required': False,
                'type': int,
                'default': 0
            },
            'max_concurrent': {
                'abbr': 'c',
                'description': 'Maximum number of trials to be run concurrently',
                'required': False,
                'type': int,
                'default': kDefaultMaxConcurrent
            },
            'stop': {
                'abbr': None,
                'description': 'Stopping criteria for the training',
                'required': False,
                'type': dict,
                'default': {'training_iteration': 1}
            },
            'scheduler_param': {
                'abbr': None,
                'description': 'Extra parameters given to the trial scheduler',
                'required': False,
                'type': (dict, type(None)),
                'default': None
            },
            'algorithm_param': {
                'abbr': None,
                'description': 'Extra parameters given to the hyperparameter optimization algorithm',
                'required': False,
                'type': (dict, type(None)),
                'default': None
            }       
        }


import json
from copy import deepcopy
from json import JSONDecodeError

from hpogrid.configuration import ConfigurationBase

class SearchSpaceConfiguration(ConfigurationBase):
    _CONFIG_TYPE_ = 'search_space'
    _CONFIG_DISPLAY_NAME_ = 'search space'
    _DESCRIPTION_ = 'Manage configuration for hyperparameter search space'
    _LIST_COLUMNS_ = ['Search Space Configuration']
    _CONFIG_FORMAT_ = \
        {
            'search_space': {
                'abbr': 's',
                'description': 'A json decodable string defining the search space',
                'required': True,
                'type': dict,
            }
        }
    
    _SEARCH_SPACE_FORMAT_ = \
        {
            'categorical': {
                'categories' : { 'type': list, 'required': True},
                'grid_search' : { 'type': int, 'required': False, 'default': 0}
            },
            'uniform': {
                'low': { 'type': (float, int), 'required': True},
                'high': { 'type': (float, int), 'required': True}
            },
            'uniformint': {
                'low': { 'type': int, 'required': True},
                'high': { 'type': int, 'required': True}
            },
            'quniform': {
                'low': { 'type': (float, int), 'required': True},
                'high': { 'type': (float, int), 'required': True},
                'q': { 'type': (float, int), 'required': True}
            },
            'loguniform': {
                'low': { 'type': (float, int), 'required': True},
                'high': { 'type': (float, int), 'required': True},
                'base': { 'type': (float, int), 'required': False}
            },
            'qloguniform': {
                'low': { 'type': (float, int), 'required': True},
                'high': { 'type': (float, int), 'required': True},
                'q': { 'type': (float, int), 'required': True},
                'base': { 'type': (float, int), 'required': False}
            },
            'normal': {
                'mu': { 'type': (float, int), 'required': True},
                'sigma': { 'type': (float, int), 'required': True}
            },
            'qnormal': {
                'mu': { 'type': (float, int), 'required': True},
                'sigma': { 'type': (float, int), 'required': True},
                'q': { 'type': (float, int), 'required': True}
            },
            'lognormal': {
                'mu': { 'type': (float, int), 'required': True},
                'sigma': { 'type': (float, int), 'required': True},
                'base': { 'type': (float, int), 'required': False}
            },
            'qlognormal': {
                'mu': { 'type': (float, int), 'required': True},
                'sigma': { 'type': (float, int), 'required': True},
                'base': { 'type': (float, int), 'required': False},
                'q': { 'type': (float, int), 'required': True}
            },
            'fixed': {
                'value': { 'type': None, 'required': True}
            }
        }
    
    @classmethod
    def validate(cls, config):
        if 'search_space' in config:
            search_space = deepcopy(config['search_space'])
        else:
            search_space = config
        # if attribute type is dict, parse string input as dict
        if isinstance(search_space, str):
            try:
                search_space = json.loads(search_space)
            except JSONDecodeError:
                raise RuntimeError('ERROR: Cannot decode the value of "search_space" as dictionary.'
                    'Please check your input.')        
        config_format = cls._SEARCH_SPACE_FORMAT_
        sampling_methods = config_format.keys()
        kHPKeys = set(['method', 'dimension'])
        for hp in search_space:
            if (not isinstance(search_space[hp], dict)) or (set(search_space[hp].keys()) != kHPKeys):
                raise ValueError('Each hyperparameter must be defined by a dictionary containing the keys "method" and "dimension"')
            if not isinstance(search_space[hp]['dimension'], dict):
                raise ValueError('Dimension of a hyperparameter must be a dictionary containing the parameters of its sampling method')
            method = search_space[hp]['method']
            # check if hyperparameter sampling method is supported
            if method not in sampling_methods:
                raise ValueError('Sampling method "{}" is not supported. '
                                 'Supported methods are: {}'.format(method,','.join(sampling_methods)))
            for arg in config_format[method]:
                if arg in search_space[hp]['dimension']:
                    # check if the argument type of the sampling method is correct
                    value_type = config_format[method][arg]['type']
                    if not isinstance(search_space[hp]['dimension'][arg], value_type):
                        raise ValueError('The value of argument "{}" for the sampling method '
                              '"{}" (for hyperparameter "{}") must be of type {}'.format(
                            arg, method, hp, type2str(value_type)))
                else:
                    # check if a required argument for the sampling method is missing
                    if (config_format[method][arg]['required']):
                        raise ValueError('Missing argument "{}" for the sampling method "{}" '
                                         'for the hyperparameter "{}"'.format(arg, method, hp))
                    # set the argument to its default value if not specified
                    if ('default' in config_format[method][arg]):
                        print('Info: The argument "{}" for the sampling method "{}" for'
                              ' the hyperparameter "{}" will be set to its default value {}'.format(
                              arg, method, hp, config_format[method][arg]['default']))
                        search_space[hp]['dimension'][arg] = config_format[method][arg]['default']

        return search_space


from hpogrid.configuration import ConfigurationBase
from hpogrid.components.defaults import kDefaultContainer, kDefaultOutDS

class GridConfiguration(ConfigurationBase):
    _CONFIG_TYPE_ = 'grid_config'
    _CONFIG_DISPLAY_NAME_ = 'grid'
    _DESCRIPTION_ = 'Manage configuration for grid job submission'
    _LIST_COLUMNS_ = ['Grid Configuration']
    _CONFIG_FORMAT_ = \
        {
            'site': {
                'abbr': 's',
                'description': 'Name of the grid site(s) to where the jobs are submitted',
                'required': False,
                'type': (list, str, type(None)),
                'default': None
            },
            'container': {
                'abbr': 'c',
                'description': 'Name of the docker or singularity container in which the jobs are run',
                'required': False,
                'type': str,
                'default': kDefaultContainer
            },
            'inDS': {
                'abbr': 'i',
                'description': 'Name of (rucio) input dataset', 
                'required': False,
                'type': (str, type(None)),
                'default': None
            },
            'outDS': {
                'abbr': 'o',
                'description': 'Name of output dataset',
                'required': False,
                'type': str,
                'default': kDefaultOutDS
            },            
            'retry': {
                'abbr': 'r',
                'description': 'Check to enable retrying faild jobs',
                'required': False,
                'type': (bool, int),
                'default': 0
            },
            'extra': {
                'abbr': 'e',
                'description': 'extra options passed to panda command',
                'required': False,
                'type': dict,
                'default': {}
            }
        }